/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/**********************************************************************
 *
 * Creation of a RMAP_3D_PROTO_SPACE structure
 *
 **********************************************************************/

RMAP_3D_PROTO_SPACE*	
rMap_create3dProtoSpace ()
{
  RMAP_3D_PROTO_SPACE* ps;
  
  ps = (RMAP_3D_PROTO_SPACE*) malloc (sizeof (RMAP_3D_PROTO_SPACE));
  if (ps == NULL) {
    rMapError(ERR_RMAP_MALLOC);
    return (NULL);
  }
  return (ps);
}

/**********************************************************************
 *
 * Destruction of a 'RMAP_3D_PROTO_SPACE'
 *
 **********************************************************************/

int
rMap_destroy3dProtoSpace (RMAP_3D_PROTO_SPACE*	ps)
{
  free (ps);
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Copy of a 'RMAP_3D_PROTO_SPACE' structure
 *
 **********************************************************************/

RMAP_3D_PROTO_SPACE*
rMap_copy3dProtoSpace (RMAP_3D_PROTO_SPACE* ps)

{
  RMAP_3D_PROTO_SPACE*	newPs;

  newPs = rMap_create3dProtoSpace ();
  if (newPs == NULL)
    return (NULL);

  memcpy (newPs, ps, sizeof (RMAP_3D_PROTO_SPACE));

  return (newPs);
}


/**********************************************************************
 *
 * Prints informations on a 'RMAP_3D_PROTO_SPACE'
 *
 **********************************************************************/

int
rMap_info3dProtoSpace (RMAP_3D_PROTO_SPACE*	ps)

{
  int	i;
  printf ("Informations on a RMAP_3D_PROTO_SPACE structure:\n");
  switch (ps->state) {
  case NIL_PROTOS: printf ("Empty proto space\n"); break;
  case FILLED_WITH_PROTOS: printf ("Probabilities not computed\n"); break;
  case PROBAS_PROTOS_OK: 
    printf ("Probabilities computed (using %d neigbors)\n", 
	    ps->nbNeighbors); break;
  }
  switch (ps->norm) {
  case NOT_COMPUTED: printf ("Attributes not computed\n"); break;
  case COMPUTED: printf ("Attributes not normalized\n"); break;
  case MINMAX_NORMALIZED: printf ("Attributes normalized (MinMax)\n"); break;
  case STATISTICALLY_NORMALIZED: printf ("Attributes normalized (statistic norm)\n"); break;
  }

  printf ("Contains %d regionMap:\n", ps->nbRegionMap);
  for (i = 0; i < ps->nbRegionMap; i++)
    printf ("\t%s\n", ps->regionMapFiles[i]);
  printf ("\n");
  printf ("Considered classes:\n");
  for (i = 1; i < NB_3D_CLASSES; i++) 
    if (ps->classes[i] != 0) 
      printf ("\t%s\n", rMap_return3dClassString(i));
  printf ("%d prototypes in total :\n", ps->nbProtos);
  for (i = 1; i < NB_3D_CLASSES; i++) 
    if (ps->classes[i] != 0) 
      printf ("\t%s: %d\n", rMap_return3dClassString(i), ps->nbPerClass[i]);
  printf ("Attributes considered:\n");
  for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
    if (ps->attributeMask[i] != 0) 
      printf ("\t%s: Min = %.3f, Max = %.3f, Delta = %.3f, Mean = %.3f, StdDev = %.3f\n",
	      rMap_return3dAttributeString(i), 
	      ps->minValues[i], ps->maxValues[i], ps->deltaValues[i],
	      ps->mean[i], sqrt(ps->var[i]));
  printf ("\n");

  return (RMAP_OK);
}

/**********************************************************************
 *
 * Prints informations on one prototype of a 'RMAP_3D_PROTO_SPACE'
 *
 **********************************************************************/

int
rMap_infoOne3dProto (RMAP_3D_PROTO_SPACE*	ps,
		     int			i)

{
  if ((i < 0) || (i >= ps->nbProtos)) {
    printf ("Enter i so that 0 <= i < %d !\n", ps->nbProtos);
    return rMapError (ERR_RMAP_INDICES);
  }
  
  if (rMap_printRegion3dInfos (&(ps->prototypes[i])) != RMAP_OK)
    return (RMAP_ERROR);
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Adds the user labeled regions of a regionMap in the protoSpace
 *
 **********************************************************************/

int
rMap_addMapTo3dProtoSpace (REGION_MAP*		rm,
			   char*		fileName,
			   RMAP_3D_PROTO_SPACE*	ps)
{
  RMAP_REGION*		region;
  double		value, mean;
  int			i, j, k;

  /* If it is the first time that prototypes are added, copy the regionMap
     classification information, and sets all the attributes */
  if (ps->state == NIL_PROTOS) {
    for (i = 1; i < NB_3D_CLASSES; i++) 
      ps->classes[i] = rm->classes3d[i];
    for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
      ps->attributeMask[i] = 1;
  }
  else { /* Check that the space has not been normalized, and also that the
	    considered classes fit */
    if ((ps->norm == MINMAX_NORMALIZED) 
	|| (ps->norm == STATISTICALLY_NORMALIZED))
      return rMapError (ERR_RMAP_PROTO_SPACE_STATUS);
    for (i = 1; i < NB_3D_CLASSES; i++) 
      if (ps->classes[i] != rm->classes3d[i])
	return rMapError (ERR_RMAP_PROTO_SPACE_STATUS);
  }
  /* Simply loop on the regions of rm, add it to the protoSpace if it has been
     labeled by the user */
  region = rm->regions;
  for (i = 0; i < rm->nblig; i++) {
    for (j = 0; j < rm->nbcol; j++) {
      if (region->infos3d.classificationState == USER_CLASSIFIED) {
	/* Compute the bounds on the attributes */
	if (ps->nbProtos == 0) 
	  for (k = 0; k < NB_3D_ATTRIBUTES; k++) {
	    ps->minValues[k] = region->infos3d.attributes[k];
	    ps->maxValues[k] = region->infos3d.attributes[k];
	  }
	else
	  for (k = 0; k < NB_3D_ATTRIBUTES; k++) {
	    ps->minValues[k] 
	      = MIN (ps->minValues[k], region->infos3d.attributes[k]);
	    ps->maxValues[k] 
	      = MAX (ps->maxValues[k], region->infos3d.attributes[k]);
	  }
	memcpy ((RMAP_REGION*) (ps->prototypes + ps->nbProtos), 
		region, sizeof(RMAP_REGION));
	ps->nbProtos += 1;
	ps->nbPerClass[region->infos3d.label] += 1;
      }
      region++;
    }
  }

  /* Loop on the prototypes to compute the attributes variation (delta), mean
     and variances */ 
  for (k = 0; k < NB_3D_ATTRIBUTES; k++) {
    value = (ps->prototypes[0]).infos3d.attributes[k];
    ps->mean[k] = value;
    ps->var[k] = 0.0;
    for (i = 1; i < ps->nbProtos; i++) {
      value = (ps->prototypes[i]).infos3d.attributes[k];
      mean = ps->mean[k];
      ps->mean[k] = ((mean * i) + value) / (i + 1);
      ps->var[k] += (value - mean) * (value - ps->mean[k]);
      ps->deltaValues[k] = ps->maxValues[k] - ps->minValues[k];
    }
  }
  /* Finish the variances computation */
  for (k = 0; k < NB_3D_ATTRIBUTES; k++) 
    ps->var[k] /= (ps->nbProtos - 1);


  strcpy (ps->regionMapFiles[ps->nbRegionMap], fileName);
  ps->nbRegionMap += 1;
  ps->state = FILLED_WITH_PROTOS;
  ps->norm = COMPUTED;
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Normalize a RMAP_3D_PROTO_SPACE structure
 *
 **********************************************************************/

int
rMap_normalize3DProtoSpace (RMAP_3D_PROTO_SPACE*	ps,
			    RMAP_ATTRIBUTES_STATE	normType)

{
  RMAP_REGION*	proto;
  int		i;

  /* Check that the protoSpace has not been normalized before, and that the
     normType is valid */ 
  if ((ps->norm == MINMAX_NORMALIZED) 
      || (ps->norm == STATISTICALLY_NORMALIZED)) {
    printf ("The prototype space has already been normalized\n");
    return rMapError (ERR_RMAP_PROTO_SPACE_STATUS);
  }
  if ((normType != MINMAX_NORMALIZED) 
      && (normType != STATISTICALLY_NORMALIZED)) {
    printf ("The norm type is not a valid one\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }
  for (i = 0; i < ps->nbProtos; i++) {
    proto = &(ps->prototypes[i]);
    rMap_normalize3DAttributes (proto, ps, normType);
  }  

  ps->norm = normType;

  return (RMAP_OK);
}

/******************************************************************
 *
 * In a prototype space whose attributes are computed and normalized, associate
 * partial probabilities to each prototype using the nearest neighbour
 * appoximation, on the attributes that are specified by the vector
 * 'attributeMask'
 *
 ******************************************************************/

int
rMap_computeProtoSpace3DProbas (RMAP_3D_PROTO_SPACE*	ps, 
				int		nbNeighbors,
				int*	 	attributeMask)

{
  int		i;
  int		class;
  RMAP_REGION*	region;

  /* Is the prototype space well set ? */
  if (ps->state != FILLED_WITH_PROTOS)
    return rMapError (ERR_RMAP_PROTO_SPACE_STATUS);
  if ((ps->norm != MINMAX_NORMALIZED) 
      && (ps->norm != STATISTICALLY_NORMALIZED))
    return rMapError (ERR_RMAP_PROTO_SPACE_STATUS);

  if (nbNeighbors <= 0) {
    printf ("The number of neighbors must be > 0 !\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }

  /* Just prints the number of neighbors that will be used */
  for (class = 0; class < NB_3D_CLASSES; class++) {
    if (ps->classes[class] != 0) {
      printf ("Class %s: nb of neighbor considered: %d\n",
	      rMap_return3dClassString (class),
	      (int) (((double) nbNeighbors) * log10 (ps->nbPerClass[class])));
    }
  }


  printf ("The counter will run up to %d\n", ps->nbProtos);
  for (i = 0; i < ps->nbProtos; i++) {
    region = &(ps->prototypes[i]);
    if (rMap_compute3DProbas (region, ps, nbNeighbors, attributeMask)
	!= RMAP_OK)
      return (RMAP_ERROR);
    /* On calcule l'entropie du pav */
/*     if (compute_3d_entropie (proto_3d) != CT_OK) */
/*       return (CF_ERROR); */
    printf ("%d\r", i);
    fflush (stdout);
  }
  printf ("\n");

  /* Copy the attribute masks, set the status properly */ 
  for (i = 0; i < NB_3D_ATTRIBUTES; i++)
    ps->attributeMask[i] = *(attributeMask + i);

  ps->state = PROBAS_PROTOS_OK;
  ps->nbNeighbors = nbNeighbors;

  return (RMAP_OK);
}

/******************************************************************
 *
 * Computes the partial probabilities of a region in a prototype space using the
 * nearest neighbour approximation. 
 *
 ******************************************************************/

int
rMap_compute3DProbas (RMAP_REGION*		region,
		      RMAP_3D_PROTO_SPACE*	ps, 
		      int			nbNeighbors,
		      int*			attributeMask)

{
  int		i, j, k, l;
  int		class;
  /* Nb of neighbours considered per class */
  int		nbNeighborsPerClass[NB_3D_CLASSES]; 

  /* Array in which the closest neighbours of 'region' will be stored */
  RMAP_REGION* 	neighbors[NB_3D_CLASSES];
  /* Tableau dans lequel seront stocks les plus
				   proches voisins de 'pave' */
  RMAP_REGION*	prototype;	/* Un des protos du proto_space */
  double	minDist;
  RMAP_REGION*	minPrototype = NULL;
  double	distance;

  double	minAttributes[NB_3D_ATTRIBUTES];
  double	maxAttributes[NB_3D_ATTRIBUTES]; 
				/* Attributs min et max d'un tableau de
				   voisins */ 

  double	volumes[NB_3D_CLASSES];
				/* Tableau dans lequel est stock le volume
				   occup par les plus proches voisins */
  double	probas[NB_3D_CLASSES];
				/* Tableau dans lequel est stocke la proba
				   partielle */
  double	somme_probas;

  /******************************************************************/
  /* Init: determination of the number of neighbors, allocation of the arrays
     tableaux 'neighbors' */
  for (class = 0; class < NB_3D_CLASSES; class++) {
    if (ps->classes[class] != 0) {
      nbNeighborsPerClass[class] 
	= (int) (((double) nbNeighbors) * log10 (ps->nbPerClass[class]));
      /* Check that there are not more neighbors than the number of prototypes
       */ 
      if (nbNeighborsPerClass[class] >= ps->nbPerClass[class]) {
	printf ("Too large number of neighbors !\n");
	return rMapError (ERR_RMAP_BAD_PARAMETER);
      }
      if (nbNeighborsPerClass[class] <= 1) {
	printf ("Too small number of neighbors !\n");
	return rMapError (ERR_RMAP_BAD_PARAMETER);
      }
      neighbors[class]
	= (RMAP_REGION*) malloc (nbNeighborsPerClass[class] 
				 * sizeof (RMAP_REGION));
      if (neighbors[class] == NULL)
	return rMapError (ERR_RMAP_MALLOC);
    }
  }

  /******************************************************************/
  /* First part: fill the arrays "neighbors" for each considered class */
  for (class = 0; class < NB_3D_CLASSES; class++) {
    if (ps->classes[class] != 0) {
      /* Put the considered region in the neighbors */
      *(neighbors[class]) = *region;
      region->updateState = RMAP_REGION_NEVER_PERCEIVED;
      /* Look for the other closest neighbors of the same class (start from 1,
	 as the current region has been considered as the closest neighbor) */
      for (i = 1; i < nbNeighborsPerClass[class]; i++) {
	minDist = FLT_MAX;	
	for (j = 0; j < ps->nbProtos; j++) {
	  prototype = &(ps->prototypes[j]);
	  if ((prototype->updateState != RMAP_REGION_NEVER_PERCEIVED) 
	      && (prototype->infos3d.label == class)) {
	    distance = 
	      rMap_compute3DProtoDistance (region, prototype, attributeMask);
	    if (distance < minDist) {
	      minDist = distance;
	      minPrototype = prototype;
	    }
	  }
	} /* End of the loop on the prototypes */
	
	if (minPrototype == NULL)
	  return rMapError (ERR_RMAP_IMPOSSIBLE_CASE);
	
	/* Put 'minPrototype' in the array of the nearest neighbors */
	*(neighbors[class] + i) = *minPrototype;
	/* Mark minPrototype to avoid considering it again, reset distMin */
	minPrototype->updateState = RMAP_REGION_NEVER_PERCEIVED;
	minDist = FLT_MAX;
      }
      /* Reset the prototypes state */
      for (j = 0; j < ps->nbProtos; j++)
	(ps->prototypes[j]).updateState = RMAP_REGION_PERCEIVED;
    }
  }

  /******************************************************************/
  /* Second part : compute the volume defined by the nearest neighbors */

  for (class = 0; class < NB_3D_CLASSES; class++) {
    if (ps->classes[class] != 0) {
      for (j = 0; j < NB_3D_ATTRIBUTES; j++) {
	minAttributes[j] = FLT_MAX; 
	maxAttributes[j] = - FLT_MAX;
      }

      /* Loop on the neigbors to find the min and max values of the
	 attributes */
      for (l = 0; l < nbNeighborsPerClass[class]; l++) {
	prototype = (neighbors[class] + l);
	for (j = 0; j < NB_3D_ATTRIBUTES; j++) {
	  if (!isnan(prototype->infos3d.attributes[j])) {
	    minAttributes[j] 
	      = MIN (minAttributes[j], prototype->infos3d.attributes[j]);
	    maxAttributes[j] 
	      = MAX (maxAttributes[j], prototype->infos3d.attributes[j]);
	  }
	}
      }
      
      /* Affectation du volume trouv */
      volumes[class] = rMap_compute3DVolume (minAttributes, 
					     maxAttributes,  
					     attributeMask);
      if (volumes[class] == 0.0)
	/* This happens if one of the min attribute value is equal to the max
	   value -- yes, it can happen */
	printf("### Warning: a volume is equal to 0.\n"
	       "### You should increase the number of neighbors considered\n");
    }
  }
  
  /******************************************************************/
  /* Third part : compute the probabilities on the basis of the found volumes
   */
  for (i = 0; i < NB_3D_CLASSES; i++) {
    if (ps->classes[i] != 0) {
      probas[i] 
	= ((double) (nbNeighborsPerClass[i]) / ((double) ps->nbPerClass[i]))
	/  volumes[i];
    }
  }
  
  /* Normalize the probas, we are done. */
  somme_probas = 0.0;
  for (i = 0; i < NB_3D_CLASSES; i++) 
    if (ps->classes[i] != 0) 
      somme_probas += probas[i];
  
  for (i = 0; i < NB_3D_CLASSES; i++)
    if (ps->classes[i] != 0) {
      region->infos3d.newProbas[i] = probas[i] / somme_probas;
      region->infos3d.mergedProbas[i] = region->infos3d.newProbas[i];
      region->infos3d.growedProbas[i] = region->infos3d.newProbas[i];
    }

  /******************************************************************/
  /* End: free the arrays of the nearest neighbors */
  for (i = 0; i < NB_3D_CLASSES; i++)
    if (ps->classes[class] != 0) 
      free (neighbors[i]);
  
  /* Update the state of the region */
  region->infos3d.probaState = RMAP_PROBAS_COMPUTED;

  return (RMAP_OK);
}


/*********************************************************************
 *
 * Returns the volume defined the min and max values of the attributes 
 *
 *********************************************************************/

double
rMap_compute3DVolume (double*    tab_min,
		      double*    tab_max,
		      int*       attributeMask)
{
  int       i;
  double volume = 1.0;
  for (i = 0; i < NB_3D_ATTRIBUTES; i++)
    if (*(attributeMask + i) != 0) {
      if (tab_max[i] == tab_min[i])
	printf ("### Attribute %s: min and max values equal\n",
		rMap_return3dAttributeString(i));
      volume = volume * (tab_max[i] - tab_min[i]);
    }
  return (volume);
}


/*********************************************************************
 *
 * Returns the distance between two regions on the basis of the attributes
 * selected by attributeMask.
 *
 ********************************************************************/

double
rMap_compute3DProtoDistance (RMAP_REGION* 	region1,
			     RMAP_REGION* 	region2,
			     int*		attributeMask)
{
  double    distance = 0.0;
  int       i;
  for (i = 0; i < NB_3D_ATTRIBUTES; i++)
    if (*(attributeMask + i) != 0)
      if (!isnan(region1->infos3d.attributes[i])
	  && !isnan(region2->infos3d.attributes[i]))
	distance += 
	  fabs (region1->infos3d.attributes[i] 
		- region2->infos3d.attributes[i]);
  return (distance);
}
